from abc import abstractmethod
from typing import List, Dict, Any

from checkov.common.checks.base_check import BaseCheck
from checkov.common.models.enums import CheckCategories, CheckResult
from checkov.terraform.checks.provider.registry import provider_registry


class BaseProviderCheck(BaseCheck):
    def __init__(self, name: str, id: str, categories: List[CheckCategories], supported_provider: List[str]) -> None:
        super().__init__(
            name=name, id=id, categories=categories, supported_entities=supported_provider, block_type="provider"
        )
        self.supported_provider = supported_provider
        provider_registry.register(self)

    def scan_entity_conf(self, conf: Dict[str, List[Any]], entity_type: str) -> CheckResult:
        return self.scan_provider_conf(conf)

    @abstractmethod
    def scan_provider_conf(self, conf: Dict[str, List[Any]]) -> CheckResult:
        raise NotImplementedError()



import logging
import os
from copy import deepcopy
from typing import Union

from checkov.common.util.consts import RESOLVED_MODULE_ENTRY_NAME
from checkov.terraform.graph_builder.graph_components.attribute_names import CustomAttributes
from checkov.terraform.graph_builder.graph_components.block_types import BlockType
from checkov.terraform.graph_builder.utils import remove_module_dependency_in_path
from checkov.terraform.checks.utils import utils
from checkov.terraform.checks.utils.utils import calculate_hash, decode_graph_property_value, join_trimmed_strings


class Block:
    def __init__(self, name, config, path, block_type, attributes, id='', source='', encode=False):
        """
            :param name: unique name given to the terraform block, for example: 'aws_vpc.example_name'
            :param config: the section in tf_definitions that belong to this block
            :param path: the file location of the block
            :param block_type: BlockType
            :param attributes: dictionary of the block's original attributes in the terraform file
        """
        self.name = name
        self.config = deepcopy(config)
        self.module_dependency = ""
        self.module_dependency_num = ""
        self.path = path
        if path:
            self.path, module_dependency, num = remove_module_dependency_in_path(path)
            self.path = os.path.realpath(self.path)
            if module_dependency:
                self.module_dependency = module_dependency
                self.module_dependency_num = num
        self.block_type = block_type
        if attributes.get(RESOLVED_MODULE_ENTRY_NAME):
            del attributes[RESOLVED_MODULE_ENTRY_NAME]
        self.attributes = attributes
        self.id = id
        self.source = source
        self.changed_attributes = {}
        self.breadcrumbs = {}
        self.module_connections = {}
        self.source_module = set()

        attributes_to_add = self._extract_inner_attributes()
        self.attributes.update(attributes_to_add)
        self.encode = encode

    def _extract_inner_attributes(self):
        attributes_to_add = {}
        for attribute_key in self.attributes:
            attribute_value = self.attributes[attribute_key]
            if type(attribute_value) is list and len(attribute_value) > 0 and type(attribute_value[0]) is dict:
                inner_attributes = get_inner_attributes(attribute_key, attribute_value)
                attributes_to_add.update(inner_attributes)
        return attributes_to_add

    def __str__(self):
        return str(self.block_type) + ': ' + self.name

    def get_attribute_dict(self):
        """
           :return: map of all the block's native attributes (from the source file),
           combined with the attributes generated by the module builder.
           If the attributes are not a primitive type, they are converted to strings.
           """
        base_attributes = self.get_base_attributes()
        self.get_origin_attributes(base_attributes)

        if self.changed_attributes:
            # add changed attributes only for calculating the hash
            changed_attributes_keys = list(self.changed_attributes.keys())
            changed_attributes_keys.sort()
            base_attributes['changed_attributes'] = changed_attributes_keys

        if self.breadcrumbs:
            sorted_breadcrumbs = dict(sorted(self.breadcrumbs.items()))
            base_attributes[CustomAttributes.RENDERING_BREADCRUMBS] = sorted_breadcrumbs

        if self.encode:
            for attribute in base_attributes:
                value_to_encode = base_attributes[attribute]
                encoded_value = utils.encode_graph_property_value(value_to_encode)
                base_attributes[attribute] = encoded_value

        base_attributes[CustomAttributes.HASH] = calculate_hash(base_attributes)

        if base_attributes.get('changed_attributes'):
            # removed changed attributes if it was added previously for calculating hash.
            del base_attributes['changed_attributes']

        return base_attributes

    def get_origin_attributes(self, base_attributes):
        for attribute_key in list(self.attributes.keys()):
            attribute_value = self.attributes[attribute_key]
            if type(attribute_value) is list and len(attribute_value) == 1:
                attribute_value = attribute_value[0]
            if type(attribute_value) is dict or type(attribute_value) is list:
                inner_attributes = get_inner_attributes(attribute_key, attribute_value)
                base_attributes.update(inner_attributes)
            if attribute_key == 'self':
                base_attributes['self_'] = attribute_value
                continue
            else:
                base_attributes[attribute_key] = attribute_value

    def get_hash(self):
        attributes_dict = self.get_attribute_dict()
        return attributes_dict.get(CustomAttributes.HASH)

    def get_decoded_attribute_dict(self):
        attributes = self.get_attribute_dict()
        if self.encode:
            for attribute_key in attributes:
                attributes[attribute_key] = decode_graph_property_value(attributes[attribute_key])
        return attributes

    def update_attribute(self, attribute_key, attribute_value, change_origin_id, previous_breadcrumbs):
        if not previous_breadcrumbs or previous_breadcrumbs[-1] != change_origin_id:
            previous_breadcrumbs.append(change_origin_id)

        self.update_inner_attribute(attribute_key, self.attributes, attribute_value)
        attribute_key_parts = attribute_key.split('.')
        if len(attribute_key_parts) == 1:
            self.attributes[attribute_key] = attribute_value
            self.changed_attributes[attribute_key] = previous_breadcrumbs
            return
        for i in range(len(attribute_key_parts)):
            key = join_trimmed_strings(char_to_join=".", str_lst=attribute_key_parts, num_to_trim=i)
            if key.find('.') > -1:
                self.attributes[key] = attribute_value
                attribute_value = {attribute_key_parts[len(attribute_key_parts)-1 - i]: attribute_value}
                self.changed_attributes[key] = previous_breadcrumbs

    def update_inner_attribute(self, attribute_key: str, nested_attributes: Union[list, dict], value_to_update):
        split_key = attribute_key.split('.')
        i = 1
        curr_key = '.'.join(split_key[0:i])
        if type(nested_attributes) is list:
            if curr_key.isnumeric():
                curr_key = int(curr_key)
                if curr_key < len(nested_attributes):
                    if not isinstance(nested_attributes[curr_key], dict):
                        nested_attributes[curr_key] = value_to_update
                    else:
                        self.update_inner_attribute('.'.join(split_key[i:]), nested_attributes[curr_key], value_to_update)
            else:
                for inner in nested_attributes:
                    self.update_inner_attribute(curr_key, inner, value_to_update)
        elif type(nested_attributes) is dict:
            while curr_key not in nested_attributes and i <= len(split_key):
                i += 1
                curr_key = '.'.join(split_key[0:i])
            if attribute_key in nested_attributes.keys():
                nested_attributes[attribute_key] = value_to_update
            if len(split_key) == 1:
                nested_attributes[curr_key] = value_to_update
            elif curr_key in nested_attributes.keys():
                self.update_inner_attribute('.'.join(split_key[i:]), nested_attributes[curr_key], value_to_update)

    def add_module_connection(self, attribute_key, vertex_id):
        if not self.module_connections.get(attribute_key):
            self.module_connections[attribute_key] = []
        self.module_connections[attribute_key].append(vertex_id)

    def find_attribute(self, attribute):
        """
        :param attribute: key to search in self.attribute
        The function searches for  attribute in self.attribute. It might not exist if the block is variable or output,
        or its search path might be different if its a resource.
        :return: the actual attribute key or None
        """
        if not attribute:
            return None

        if self.attributes.get(attribute[0]):
            return attribute[0]

        if self.block_type == BlockType.VARIABLE:
            return 'default' if self.attributes.get('default') else None

        if self.block_type == BlockType.OUTPUT:
            return 'value' if self.attributes.get('value') else None

        if self.block_type == BlockType.RESOURCE and len(attribute) > 1:
            # handle cases where attribute_at_dest == ['aws_s3_bucket.template_bucket', 'acl']
            if self.name == attribute[0] and self.attributes.get(attribute[1]):
                return attribute[1]

        return None

    def get_export_data(self):
        return {'type': self.block_type.value, 'name': self.name, 'path': self.path}

    def get_base_attributes(self):
        return {
            CustomAttributes.BLOCK_NAME: self.name,
            CustomAttributes.BLOCK_TYPE: self.block_type.value,
            CustomAttributes.FILE_PATH: self.path,
            CustomAttributes.CONFIG: self.config,
            CustomAttributes.LABEL: self.__str__(),
            CustomAttributes.ID: self.id,
            CustomAttributes.SOURCE: self.source
        }


def get_inner_attributes(attribute_key, attribute_value):
    inner_attributes = {}
    if type(attribute_value) is list and len(attribute_value) == 1:
        attribute_value = attribute_value[0]

    if type(attribute_value) in [dict, list]:
        inner_attributes[attribute_key] = [None] * len(attribute_value) if type(attribute_value) is list else {}
        iterator = range(len(attribute_value)) if type(attribute_value) is list else list(attribute_value.keys())
        for key in iterator:
            if key != '':
                inner_key = f'{attribute_key}.{key}'
                inner_value = attribute_value[key]
                inner_attributes.update(get_inner_attributes(inner_key, inner_value))
                inner_attributes[attribute_key][key] = inner_attributes[inner_key]
            else:
                del attribute_value[key]
    else:
        inner_attributes[attribute_key] = attribute_value
    return inner_attributes



from checkov.common.models.enums import CheckResult, CheckCategories
from checkov.cloudformation.checks.resource.base_resource_check import BaseResourceCheck


class GlueDataCatalogEncryption(BaseResourceCheck):
    def __init__(self):
        name = "Ensure Glue Data Catalog Encryption is enabled"
        id = "CKV_AWS_94"
        supported_resources = ['AWS::Glue::DataCatalogEncryptionSettings']
        categories = [CheckCategories.LOGGING]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf):
        connection_encrypted = False
        encrypted_at_rest = False
        if 'Properties' in conf.keys():
            if 'DataCatalogEncryptionSettings' in conf['Properties'].keys():
                dc_enc_settings = conf['Properties']['DataCatalogEncryptionSettings']
                if 'ConnectionPasswordEncryption' in dc_enc_settings.keys():
                    con_pass_enc = dc_enc_settings['ConnectionPasswordEncryption']
                    if 'ReturnConnectionPasswordEncrypted' in con_pass_enc.keys():
                        if con_pass_enc['ReturnConnectionPasswordEncrypted'] == True:
                            connection_encrypted = True

                if 'EncryptionAtRest' in dc_enc_settings.keys():
                    enc_at_rest = dc_enc_settings['EncryptionAtRest']
                    if 'CatalogEncryptionMode' in enc_at_rest.keys():
                        if enc_at_rest['CatalogEncryptionMode'] == "SSE-KMS":
                            encrypted_at_rest = True

        if connection_encrypted and encrypted_at_rest:
            return CheckResult.PASSED
        
        return CheckResult.FAILED


check = GlueDataCatalogEncryption()



import logging
import os
from copy import deepcopy

from checkov.common.graph.graph_builder import reserved_attribute_names, EncryptionValues
from checkov.common.graph.graph_builder import Edge
from checkov.terraform.checks.utils.dependency_path_handler import unify_dependency_path
from checkov.terraform.graph_builder.graph_components.attribute_names import CustomAttributes
from checkov.terraform.graph_builder.graph_components.block_types import BlockType
from checkov.terraform.graph_builder.graph_components.generic_resource_encryption import \
    ENCRYPTION_BY_RESOURCE_TYPE
from checkov.terraform.graph_builder.utils import is_local_path
from checkov.terraform.checks.utils.utils import get_referenced_vertices_in_value, update_dictionary_attribute, \
    join_trimmed_strings, filter_sub_keys, attribute_has_nested_attributes
from checkov.terraform.checks.utils.utils import remove_index_pattern_from_str, calculate_hash
from checkov.terraform.variable_rendering.renderer import VariableRenderer

MODULE_RESERVED_ATTRIBUTES = ('source', 'version')


class LocalGraph:
    def __init__(self, module, module_dependency_map):
        self.module = module
        self.vertices = []
        self.edges = []
        self.in_edges = {}  # map between vertex index and the edges entering it
        self.out_edges = {}  # map between vertex index and the edges exiting it
        self.vertices_by_block_type = {}
        self.vertex_hash_cache = {}
        self.vertices_block_name_map = {}
        self.module_dependency_map = module_dependency_map
        self.map_path_to_module = {}
        self.relative_paths_cache = {}

    def build_graph(self, render_variables):
        self._create_vertices()
        undetermined_values = self._set_variables_values_from_modules()
        self._build_edges()
        self.calculate_encryption_attribute()
        if render_variables:
            logging.info('Rendering variables')
            renderer = VariableRenderer(self)
            renderer.render_variables_from_local_graph()
            self.update_vertices_breadcrumbs_and_module_connections()
            self.process_undetermined_values(undetermined_values)

    def _create_vertices(self):
        logging.info('Creating vertices')
        self.vertices = [None] * len(self.module.blocks)
        for i, block in enumerate(self.module.blocks):
            self.vertices[i] = block

            if not self.vertices_by_block_type.get(block.block_type):
                self.vertices_by_block_type[block.block_type] = []
            self.vertices_by_block_type[block.block_type].append(i)

            if block.block_type not in self.vertices_block_name_map:
                self.vertices_block_name_map[block.block_type] = {}
            if block.name not in self.vertices_block_name_map[block.block_type]:
                self.vertices_block_name_map[block.block_type][block.name] = []
            self.vertices_block_name_map[block.block_type][block.name].append(i)

            if block.block_type == BlockType.MODULE:
                # map between file paths and module vertices indexes from that file
                if not self.map_path_to_module.get(block.path):
                    self.map_path_to_module[block.path] = []
                self.map_path_to_module[block.path].append(i)

            self.in_edges[i] = []
            self.out_edges[i] = []

    def _set_variables_values_from_modules(self):
        undetermined_values = []
        for module_vertex_id in self.vertices_by_block_type.get(BlockType.MODULE, []):
            module_vertex = self.vertices[module_vertex_id]
            for attribute_name in module_vertex.attributes:
                matching_variables = self.vertices_block_name_map.get(BlockType.VARIABLE, {}).get(attribute_name, [])
                for variable_vertex_id in matching_variables:
                    variable_vertex = self.vertices[variable_vertex_id]
                    variable_dir = os.path.dirname(variable_vertex.path)
                    if module_vertex.path in self.module_dependency_map.get(variable_dir, []):
                        attribute_value = module_vertex.attributes[attribute_name]
                        has_var_reference = get_referenced_vertices_in_value(value=attribute_value, aliases={},
                                                            resources_types=self.get_resources_types_in_graph())
                        if has_var_reference:
                            undetermined_values.append(
                                {'module_vertex_id': module_vertex_id, 'attribute_name': attribute_name,
                                 'variable_vertex_id': variable_vertex_id})
                        var_default_value = self.vertices[variable_vertex_id].attributes.get("default")
                        if not has_var_reference or not var_default_value or get_referenced_vertices_in_value(value=var_default_value, aliases={},
                                                            resources_types=self.get_resources_types_in_graph()):
                            self.update_vertex_attribute(variable_vertex_id, 'default', attribute_value,
                                                         module_vertex_id, attribute_name)
        return undetermined_values

    def process_undetermined_values(self, undetermined_values):
        for undetermined in undetermined_values:
            module_vertex = self.vertices[undetermined.get('module_vertex_id')]
            value = module_vertex.attributes.get(undetermined.get('attribute_name'))
            if not get_referenced_vertices_in_value(value=value, aliases={},
                                                    resources_types=self.get_resources_types_in_graph()):
                self.update_vertex_attribute(undetermined.get('variable_vertex_id'), 'default', value,
                                             undetermined.get('module_vertex_id'), undetermined.get('attribute_name'))

    def _get_aliases(self):
        """
        :return aliases: map between alias names that are found inside the blocks and the block type their aliased to.
        """
        aliases = {}
        for vertex in self.vertices:
            if 'alias' in vertex.attributes:
                aliases[vertex.name] = {CustomAttributes.BLOCK_TYPE: vertex.block_type}
        return aliases

    def get_module_vertices_mapping(self):
        """
        For each vertex, if it's originated in a module import, add to the vertex the index of the
        matching module vertex as 'source_module'
        """
        block_dirs_to_modules = {}
        for dir_name, paths_to_modules in self.module_dependency_map.items():
            # for each directory, find the module vertex that imported it
            if block_dirs_to_modules.get(dir_name):
                continue
            for path_to_module in paths_to_modules:
                if not path_to_module:
                    continue
                path_to_module = unify_dependency_path(path_to_module)
                module_list = self.map_path_to_module.get(path_to_module, [])
                for module_index in module_list:
                    module_vertex = self.vertices[module_index]
                    module_vertex_dir = os.path.dirname(module_vertex.path)
                    module_source = self.vertices[module_index].attributes.get('source', [''])[0]
                    if self._get_dest_module_path(module_vertex_dir, module_source) == dir_name:
                        if not block_dirs_to_modules.get(dir_name):
                            block_dirs_to_modules[dir_name] = set()
                        block_dirs_to_modules[dir_name].add(module_index)

        for vertex in self.vertices:
            # match the right module vertex according to the vertex path directory
            module_index = block_dirs_to_modules.get(os.path.dirname(vertex.path), set())
            if module_index:
                vertex.source_module = module_index

    def _build_edges(self):
        logging.info('Creating edges')
        self.get_module_vertices_mapping()
        aliases = self._get_aliases()
        for origin_node_index, vertex in enumerate(self.vertices):
            for attribute_key in vertex.attributes:
                if attribute_key in reserved_attribute_names or attribute_has_nested_attributes(attribute_key,
                                                                                                     vertex.attributes):
                    continue
                referenced_vertices = get_referenced_vertices_in_value(value=vertex.attributes[attribute_key],
                                                                       aliases=aliases,
                                                                       resources_types=self.get_resources_types_in_graph())
                for vertex_reference in referenced_vertices:
                    # for certain blocks such as data and resource, the block name is composed from several parts.
                    # the purpose of the loop is to avoid not finding the node if the name has several parts
                    sub_values = [remove_index_pattern_from_str(sub_value) for sub_value in vertex_reference.sub_parts]
                    for i in range(len(sub_values)):
                        reference_name = join_trimmed_strings(char_to_join=".", str_lst=sub_values, num_to_trim=i)
                        if vertex.module_dependency:
                            dest_node_index = self._find_vertex_index_relative_to_path(vertex_reference.block_type,
                                                                                       reference_name,
                                                                                       vertex.path,
                                                                                       vertex.module_dependency)
                            if dest_node_index == -1:
                                dest_node_index = self._find_vertex_index_relative_to_path(vertex_reference.block_type,
                                                                                           reference_name,
                                                                                           vertex.path,
                                                                                           vertex.path)
                        else:
                            dest_node_index = self._find_vertex_index_relative_to_path(vertex_reference.block_type,
                                                                                       reference_name,
                                                                                       vertex.path,
                                                                                       vertex.module_dependency)
                        if dest_node_index > -1 and origin_node_index > -1:
                            if vertex_reference.block_type == BlockType.MODULE:
                                try:
                                    self._connect_module(sub_values, attribute_key, self.vertices[dest_node_index],
                                                         origin_node_index)
                                except Exception as e:
                                    logging.warning(
                                        f'Module {self.vertices[dest_node_index]} does not have source attribute, skipping')
                                    logging.warning(e, stack_info=True)
                            else:
                                self._create_edge(origin_node_index, dest_node_index, attribute_key)
                            break

            if vertex.block_type == BlockType.MODULE:
                target_path = vertex.path
                if vertex.module_dependency != '':
                    target_path = unify_dependency_path([vertex.module_dependency, vertex.path])
                target_variables = list(filter(lambda v: self.vertices[v].module_dependency == target_path,
                                               self.vertices_by_block_type.get(BlockType.VARIABLE, {})))
                for attribute, value in vertex.attributes.items():
                    if attribute in MODULE_RESERVED_ATTRIBUTES:
                        continue
                    target_variable = None
                    for v in target_variables:
                        if self.vertices[v].name == attribute:
                            target_variable = v
                            break
                    if target_variable is not None:
                        self._create_edge(target_variable, origin_node_index, 'default')
            elif vertex.block_type == BlockType.TF_VARIABLE:
                if vertex.module_dependency != '':
                    target_path = unify_dependency_path([vertex.module_dependency, vertex.path])
                # Assuming the tfvars file is in the same directory as the variables file (best practice)
                target_variables = list(filter(lambda v: os.path.dirname(self.vertices[v].path) == os.path.dirname(vertex.path),
                                               self.vertices_block_name_map.get(BlockType.VARIABLE, {}).get(vertex.name, [])))
                if len(target_variables) == 1:
                    self._create_edge(target_variables[0], origin_node_index, 'default')


    def _create_edge(self, origin_vertex_index, dest_vertex_index, label):
        if origin_vertex_index == dest_vertex_index:
            return
        edge = Edge(origin_vertex_index, dest_vertex_index, label)
        self.edges.append(edge)
        self.out_edges[origin_vertex_index].append(edge)
        self.in_edges[dest_vertex_index].append(edge)

    def _connect_module(self, sub_values, attribute_key, module_node, origin_node_index):
        """
        :param sub_values: list of sub values of the attribute value.
                            example: given 'instance_type = module.child.myoutput',
                                        then attribute_key = instance_type, sub_values = ['child', 'myoutput']
        :param attribute_key: the name of the attribute that has module block as value
        :param module_node: the graph node of the module

        The function receives a node of a block of type BlockType.Module, and finds all the nodes of blocks that belong to this
        module, and creates edges between them.
        """
        curr_module_dir = os.path.dirname(module_node.path)
        dest_module_source = module_node.attributes['source'][0]
        dest_module_path = self._get_dest_module_path(curr_module_dir, dest_module_source)

        if len(sub_values) > 1:
            block_name_in_other_module = sub_values[1]
            output_blocks_with_name = self.vertices_block_name_map.get(BlockType.OUTPUT, {}).get(
                block_name_in_other_module, [])
            for vertex_index in output_blocks_with_name:
                vertex = self.vertices[vertex_index]
                if os.path.dirname(vertex.path) == dest_module_path:
                    self._create_edge(origin_node_index, vertex_index, attribute_key)
                    self.vertices[origin_node_index].add_module_connection(attribute_key, vertex_index)
                    break

    def _get_dest_module_path(self, curr_module_dir, dest_module_source):
        """
        :param curr_module_dir: current source directory
        :param dest_module_source: the value of module.source
        :return: the real path in the local file system of the dest module
        """
        dest_module_path = ''
        if is_local_path(curr_module_dir, dest_module_source):
            dest_module_path = os.path.join(curr_module_dir, dest_module_source)
        else:
            for root, d_names, f_names in os.walk(self.module.source_dir):
                dest_module_path = os.path.join(root, dest_module_source)
                if os.path.exists(dest_module_path):
                    break
        return os.path.realpath(dest_module_path)

    def _find_vertex_index_relative_to_path(self, block_type, name, block_path, module_path):
        relative_vertices = []
        possible_vertices = self.vertices_block_name_map.get(block_type, {}).get(name, [])
        for vertex_index in possible_vertices:
            vertex = self.vertices[vertex_index]
            if vertex.module_dependency == module_path and os.path.dirname(vertex.path) == os.path.dirname(block_path):
                relative_vertices.append(vertex_index)

        if len(relative_vertices) == 1:
            return relative_vertices[0]
        return self._find_vertex_with_longest_path_match(relative_vertices, block_path)

    def _find_vertex_with_longest_path_match(self, relevant_vertices_indexes, origin_path) -> int:
        vertex_index_with_longest_common_prefix = -1
        longest_common_prefix = ''
        for vertex_index in relevant_vertices_indexes:
            vertex = self.vertices[vertex_index]
            common_prefix = os.path.commonpath([os.path.realpath(vertex.path), os.path.realpath(origin_path)])
            if len(common_prefix) > len(longest_common_prefix):
                vertex_index_with_longest_common_prefix = vertex_index
                longest_common_prefix = common_prefix
        return vertex_index_with_longest_common_prefix

    def get_vertices_hash_codes_to_attributes_map(self):
        return {vertex.get_hash(): vertex.get_decoded_attribute_dict() for vertex in self.vertices}

    def order_edges_by_hash_codes(self):
        edges = {}
        for edge in self.edges:
            edge_data = {'edge_label': edge.label,
                         'from_vertex_hash': self.get_vertex_hash_by_index(vertex_index=edge.origin),
                         'to_vertex_hash': self.get_vertex_hash_by_index(vertex_index=edge.dest),
                         }
            edge_hash = calculate_hash(edge_data)
            edges[edge_hash] = edge
        return edges

    def get_vertex_attributes_by_index(self, index):
        return self.vertices[index].get_decoded_attribute_dict()

    def get_vertices_with_degrees_conditions(self, out_degree_cond, in_degree_cond):
        vertices_with_out_degree = set([vertex_index for vertex_index in self.out_edges.keys() if
                                        out_degree_cond(len(self.out_edges.get(vertex_index)))])
        vertices_with_in_degree = set([vertex_index for vertex_index in self.in_edges.keys() if
                                       in_degree_cond(len(self.in_edges.get(vertex_index)))])

        return list(vertices_with_in_degree.intersection(vertices_with_out_degree))

    def get_vertex_hash_by_index(self, vertex_index):
        if vertex_index not in self.vertex_hash_cache:
            self.vertex_hash_cache[vertex_index] = self.vertices[vertex_index].get_hash()
        return self.vertex_hash_cache[vertex_index]

    def get_in_edges(self, end_vertices):
        res = []
        for vertex in end_vertices:
            res.extend(self.in_edges.get(vertex, []))
        return self.sort_edged_by_dest_out_degree(res)

    def sort_edged_by_dest_out_degree(self, edges):
        edged_by_out_degree = {}
        for edge in edges:
            dest_out_degree = len(self.out_edges.get(edge.dest))
            if not edged_by_out_degree.get(dest_out_degree):
                edged_by_out_degree[dest_out_degree] = []
            edged_by_out_degree[dest_out_degree].append(edge)
        sorted_edges = []
        for degree in sorted(edged_by_out_degree.keys()):
            sorted_edges.extend(edged_by_out_degree[degree])
        return sorted_edges

    def update_vertex_attribute(self, vertex_index, attribute_key, attribute_value, change_origin_id,
                                attribute_at_dest):
        previous_breadcrumbs = []
        attribute_at_dest = self.vertices[change_origin_id].find_attribute(attribute_at_dest)
        if attribute_at_dest:
            previous_breadcrumbs = self.vertices[change_origin_id].changed_attributes.get(attribute_at_dest, [])
        self.vertices[vertex_index].update_attribute(attribute_key, attribute_value, change_origin_id,
                                                     previous_breadcrumbs)

    def update_vertices_configs(self):
        for vertex in self.vertices:
            changed_attributes = list(vertex.changed_attributes.keys())
            changed_attributes = filter_sub_keys(changed_attributes)
            self.update_vertex_config(vertex, changed_attributes)

    @staticmethod
    def update_vertex_config(vertex, changed_attributes):
        updated_config = deepcopy(vertex.config)
        if vertex.block_type != BlockType.LOCALS:
            parts = vertex.name.split('.')
            start = 0
            end = 1
            while end <= len(parts):
                cur_key = '.'.join(parts[start:end])
                if cur_key in updated_config:
                    updated_config = updated_config[cur_key]
                    start = end
                end += 1
        for changed_attribute in changed_attributes:
            new_value = vertex.attributes.get(changed_attribute, None)
            if new_value is not None:
                if vertex.block_type == BlockType.LOCALS:
                    changed_attribute = changed_attribute.replace(vertex.name + ".", '')
                updated_config = update_dictionary_attribute(updated_config, changed_attribute, new_value)

        if len(changed_attributes) > 0:
            if vertex.block_type == BlockType.LOCALS:
                updated_config = updated_config.get(vertex.name)
            update_dictionary_attribute(vertex.config, vertex.name, updated_config)

    def get_resources_types_in_graph(self):
        return self.module.get_resources_types()

    def update_vertices_breadcrumbs_and_module_connections(self):
        """
        The function processes each vertex's breadcrumbs:
        1. Get more data to each vertex in breadcrumb (name, path, hash and type)
        2. If a breadcrumb is originated in a different module, it will have 'module_connection'=True
        3. If a vertex has a 'source module' we will add a special breadcrumb for it
        """
        for vertex in self.vertices:
            for attribute_key, breadcrumbs_list in vertex.changed_attributes.items():
                hash_breadcrumbs = []
                for vertex_id in breadcrumbs_list:
                    v = self.vertices[vertex_id]
                    breadcrumb = v.get_export_data()
                    breadcrumb['module_connection'] = self._determine_if_module_connection(breadcrumbs_list, v)
                    hash_breadcrumbs.append(breadcrumb)
                vertex.breadcrumbs[attribute_key] = hash_breadcrumbs
            if len(vertex.source_module) == 1:
                m = self.vertices[list(vertex.source_module)[0]]
                source_module_data = [m.get_export_data()]
                while len(m.source_module) == 1:
                    m = self.vertices[list(m.source_module)[0]]
                    source_module_data.append(m.get_export_data())
                source_module_data.reverse()
                vertex.breadcrumbs[CustomAttributes.SOURCE_MODULE] = source_module_data

    @staticmethod
    def _determine_if_module_connection(breadcrumbs_list, vertex_in_breadcrumbs):
        """
        :param breadcrumbs_list: list of vertex's breadcrumbs
        :param vertex_in_breadcrumbs: one of the vertices in the breadcrumb list
        :return: True if vertex_in_breadcrumbs has in its module_connections at least one of the vertices in breadcrumbs_list
        """
        if not vertex_in_breadcrumbs.module_connections:
            return False
        for connection_list in vertex_in_breadcrumbs.module_connections.values():
            same_ids = [i for i in connection_list if i in breadcrumbs_list]
            if len(same_ids) > 0:
                return True
        return False

    def calculate_encryption_attribute(self):
        for vertex_index in self.vertices_by_block_type.get(BlockType.RESOURCE.value, []):
            vertex = self.vertices[vertex_index]
            resource_type = vertex.id.split('.')[0]
            encryption_conf = ENCRYPTION_BY_RESOURCE_TYPE.get(resource_type)
            attributes = vertex.get_attribute_dict()
            if encryption_conf:
                is_encrypted, reason = encryption_conf.is_encrypted(attributes)
                # TODO: Does not support possible dependency (i.e. S3 Object being encrypted due to S3 Bucket config)
                vertex.attributes[CustomAttributes.ENCRYPTION] = \
                    EncryptionValues.ENCRYPTED.value if is_encrypted else EncryptionValues.UNENCRYPTED.value
                vertex.attributes[CustomAttributes.ENCRYPTION_DETAILS] = reason



from checkov.common.models.enums import CheckResult, CheckCategories
from checkov.arm.base_resource_check import BaseResourceCheck

class AKSApiServerAuthorizedIpRanges(BaseResourceCheck):
    def __init__(self):
        # apiVersion 2017-08-03 and 2018-03-31 = Fail - No authorized IP range available
        # apiVersion 2019-02-01, 2019-04-01, 2019-06-01 - Preview
        # apiversion 2019-08-01 and greater are fully supported
        name = "Ensure AKS has an API Server Authorized IP Ranges enabled"
        id = "CKV_AZURE_6"
        supported_resources = ['Microsoft.ContainerService/managedClusters']
        categories = [CheckCategories.KUBERNETES]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf):
        if "apiVersion" in conf:
            if conf["apiVersion"] in ["2017-08-31", "2018-03-31"]:
                # ApiServerAuthorizedIpRanges not supported in these API versions
                return CheckResult.FAILED
            elif conf["apiVersion"] in ["2019-02-01", "2019-04-01", "2019-06-01"]:
                # apiServerAuthorizedIPRanges in Preview in these API versions
                if "properties" in conf:
                    if "apiServerAuthorizedIPRanges" in conf["properties"]:
                        if conf["properties"]["apiServerAuthorizedIPRanges"]:
                            return CheckResult.PASSED
            else:
                # ApiServerAuthorizedIpRanges fully supported in all future API versions
                if "properties" in conf:
                    if "apiServerAccessProfile" in conf["properties"]:
                        if "authorizedIPRanges" in conf["properties"]["apiServerAccessProfile"]:
                            if conf["properties"]["apiServerAccessProfile"]["authorizedIPRanges"]:
                                return CheckResult.PASSED
        return CheckResult.FAILED

check = AKSApiServerAuthorizedIpRanges()


from checkov.common.models.enums import CheckCategories
from checkov.terraform.checks.resource.base_resource_negative_value_check import BaseResourceNegativeValueCheck


class S3PublicACLWrite(BaseResourceNegativeValueCheck):
    def __init__(self):
        name = "S3 Bucket has an ACL defined which allows public WRITE access."
        id = "CKV_AWS_57"
        supported_resources = ['aws_s3_bucket']
        categories = [CheckCategories.GENERAL_SECURITY]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def get_inspected_key(self):
        return 'acl'

    def get_forbidden_values(self):
        return ["public-read-write"]


check = S3PublicACLWrite()



from checkov.common.models.enums import CheckResult, CheckCategories
from checkov.terraform.checks.resource.base_resource_value_check import BaseResourceCheck


class VMStorageOsDisk(BaseResourceCheck):
    def __init__(self):
        name = "Ensure that Virtual Machines use managed disks"
        id = "CKV_AZURE_92"
        supported_resources = ['azurerm_linux_virtual_machine', 'azurerm_windows_virtual_machine']
        categories = [CheckCategories.GENERAL_SECURITY]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf):
        storage_os_disk = conf.get('storage_os_disk')
        storage_data_disk = conf.get('storage_data_disk')
        if storage_os_disk and 'vhd_uri' in storage_os_disk[0]:
            return CheckResult.FAILED
        if storage_data_disk and 'vhd_uri' in storage_data_disk[0]:
            return CheckResult.FAILED
        return CheckResult.PASSED


check = VMStorageOsDisk()



from checkov.common.models.enums import CheckCategories, CheckResult
from checkov.terraform.checks.resource.base_resource_check import BaseResourceCheck


class FrontdoorUseWAFMode(BaseResourceCheck):
    def __init__(self):
        name = "Ensure that Azure Front Door uses WAF in \"Detection\" or \"Prevention\" modes"
        id = "CKV_AZURE_123"
        supported_resources = ['azurerm_frontdoor_firewall_policy']
        categories = [CheckCategories.NETWORKING]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf):
        if 'policy_settings' in conf and conf['policy_settings'][0]:
            policy_settings = conf['policy_settings'][0]
            if 'enabled' in policy_settings and not policy_settings['enabled'][0]:
                return CheckResult.FAILED
        return CheckResult.PASSED


check = FrontdoorUseWAFMode()



from checkov.cloudformation.checks.resource.base_resource_check import BaseResourceCheck
from checkov.cloudformation.parser.node import dict_node
from checkov.common.models.enums import CheckResult, CheckCategories


class DocDBAuditLogs(BaseResourceCheck):
    def __init__(self) -> None:
        name = "Ensure DocDB has audit logs enabled"
        id = "CKV_AWS_104"
        supported_resources = ["AWS::DocDB::DBClusterParameterGroup"]
        categories = [CheckCategories.LOGGING]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf: dict_node) -> CheckResult:
        params = conf.get("Properties", {}).get("Parameters", {})

        if params.get("audit_logs") == "enabled":
            return CheckResult.PASSED

        return CheckResult.FAILED


check = DocDBAuditLogs()



import io
import logging
import operator
import os
import subprocess #nosec
import tempfile
from functools import reduce
import shutil
import json

from checkov.common.output.record import Record
from checkov.common.output.report import Report
from checkov.common.runners.base_runner import BaseRunner, filter_ignored_directories
from checkov.kubernetes.runner import Runner as k8_runner
from checkov.helm.registry import registry
from checkov.runner_filter import RunnerFilter

import yaml
import asyncio

K8_POSSIBLE_ENDINGS = [".yaml", ".yml", ".json"]


class Runner(BaseRunner):
    check_type = "helm"
    helm_command = 'helm'
    system_deps = True

    @staticmethod
    def find_chart_directories(root_folder, files):
        chart_directories = []
        if files:
            logging.info('Running with --file argument; checking for Helm Chart.yaml files')
            for file in files:
                if os.path.basename(file) == 'Chart.yaml':
                    chart_directories.append(os.path.dirname(file))

        if root_folder:
            for root, d_names, f_names in os.walk(root_folder):
                filter_ignored_directories(d_names)
                if 'Chart.yaml' in f_names:
                    chart_directories.append(root)

        return chart_directories

    @staticmethod
    def parse_helm_dependency_output(o):
        output = o.decode('utf-8')
        chart_dependencies={}
        if "WARNING" in output:
            #Helm  output showing no deps, example: 'WARNING: no dependencies at helm-charts/charts/prometheus-kafka-exporter/charts\n'
            pass
        else: 
            lines = output.split('\n')
            for line in lines:
                if line != "":
                    if not "NAME" in line:
                        chart_name, chart_version, chart_repo, chart_status = line.split("\t")
                        chart_dependencies.update({chart_name.rstrip():{'chart_name': chart_name.rstrip(), 'chart_version': chart_version.rstrip(), 'chart_repo': chart_repo.rstrip(), 'chart_status': chart_status.rstrip()}})
        return chart_dependencies

    @staticmethod
    def parse_helm_chart_details(chart_path):
        with open(f"{chart_path}/Chart.yaml", 'r') as chartyaml:
            try:
                chart_meta = yaml.safe_load(chartyaml)
            except yaml.YAMLError as exc:
                logging.info(f"Failed to load chart metadata from {chart_path}/Chart.yaml. details: {exc}")
        return chart_meta

    def check_system_deps(self):
        # Ensure local system dependancies are available and of the correct version.
        # Returns framework names to skip if deps fail.
        logging.info(f"Checking necessary system dependancies for {self.check_type} checks.")
        try:
            proc = subprocess.Popen([self.helm_command, 'version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE) #nosec
            o, e = proc.communicate()
            oString = str(o, 'utf-8')
            if "Version:" in oString:
                helmVersionOutput = oString[oString.find(':')+2 : oString.find(',')-1]
                if "v3" in helmVersionOutput:
                    logging.info(f"Found working version of {self.check_type} dependancies: {helmVersionOutput}")
                    return None
            else:
                return self.check_type
        except Exception:
            logging.info(f"Error running necessary tools to process {self.check_type} checks.")
            return self.check_type

    def run(self, root_folder, external_checks_dir=None, files=None, runner_filter=RunnerFilter(), collect_skip_comments=True):

        definitions = {}
        definitions_raw = {}
        parsing_errors = {}
        files_list = []
        if external_checks_dir:
            for directory in external_checks_dir:
                registry.load_external_checks(directory)

        chart_directories = self.find_chart_directories(root_folder, files)

        report = Report(self.check_type)
    
        for chart_dir in chart_directories:
            #chart_name = os.path.basename(chart_dir)
            chart_meta = self.parse_helm_chart_details(chart_dir)
            logging.info(f"Processing chart found at: {chart_dir}, name: {chart_meta['name']}, version: {chart_meta['version']}")
            with tempfile.TemporaryDirectory() as target_dir:
                #dependency list is nicer to parse than dependency update.
                proc = subprocess.Popen([self.helm_command, 'dependency', 'list' , chart_dir], stdout=subprocess.PIPE, stderr=subprocess.PIPE) #nosec
                o, e = proc.communicate()
                if e:
                    if "Warning: Dependencies" in str(e, 'utf-8'):
                        logging.info(f"V1 API chart without Chart.yaml dependancies. Skipping chart dependancy list for {chart_meta['name']} at dir: {chart_dir}. Working dir: {target_dir}. Error details: {str(e, 'utf-8')}")       
                    else: 
                        logging.info(f"Error processing helm dependancies for {chart_meta['name']} at source dir: {chart_dir}. Working dir: {target_dir}. Error details: {str(e, 'utf-8')}")
                        
                self.parse_helm_dependency_output(o)

                try:
                    #--dependency-update needed to pull in deps before templating.
                    proc = subprocess.Popen([self.helm_command, 'template', '--dependency-update', chart_dir], stdout=subprocess.PIPE, stderr=subprocess.PIPE) #nosec
                    o, e = proc.communicate()
                    logging.debug(f"Ran helm command to template chart output. Chart: {chart_meta['name']}. dir: {target_dir}. Output: {str(o, 'utf-8')}")

                except Exception:
                    logging.info(f"Error processing helm chart {chart_meta['name']} at dir: {chart_dir}. Working dir: {target_dir}. Error details: {str(e, 'utf-8')}")
            
                output = str(o, 'utf-8')
                reader = io.StringIO(output)
                cur_source_file = None
                cur_writer = None
                last_line_dashes = False
                line_num = 1
                for s in reader:
                    s = s.rstrip()
                    if s == '---':
                        last_line_dashes = True
                        continue

                    if last_line_dashes:
                        # The next line should contain a "Source" comment saying the name of the file it came from
                        # So we will close the old file, open a new file, and write the dashes from last iteration plus this line

                        if not s.startswith('# Source: '):
                            raise Exception(f'Line {line_num}: Expected line to start with # Source: {s}')
                        source = s[10:]
                        if source != cur_source_file:
                            if cur_writer:
                                cur_writer.close()
                            file_path = os.path.join(target_dir, source)
                            parent = os.path.dirname(file_path)
                            os.makedirs(parent, exist_ok=True)
                            cur_source_file = source
                            cur_writer = open(os.path.join(target_dir, source), 'a')
                        cur_writer.write('---' + os.linesep)
                        cur_writer.write(s + os.linesep)

                        last_line_dashes = False
                    else:
                        if s.startswith('# Source: '):
                            raise Exception(f'Line {line_num}: Unexpected line starting with # Source: {s}')

                        if not cur_writer:
                            continue
                        else:
                            cur_writer.write(s + os.linesep)

                    line_num += 1

                if cur_writer:
                    cur_writer.close()

                try:
                    k8s_runner = k8_runner()
                    chart_results = k8s_runner.run(target_dir, external_checks_dir=external_checks_dir, runner_filter=runner_filter, helmChart=chart_meta['name'])
                    logging.debug(f"Sucessfully ran k8s scan on {chart_meta['name']}. Scan dir : {target_dir}")
                    report.failed_checks += chart_results.failed_checks
                    report.passed_checks += chart_results.passed_checks
                    report.parsing_errors += chart_results.parsing_errors
                    report.skipped_checks += chart_results.skipped_checks

                except:
                    with tempfile.TemporaryDirectory() as save_error_dir:
                        logging.debug(f"Error running k8s scan on {chart_meta['name']}. Scan dir: {target_dir}. Saved context dir: {save_error_dir}")
                        shutil.move(target_dir, save_error_dir) 
                    
        
        ## TODO: Export helm dependancies for the chart we've extracted in chart_dependencies
        return report


    def _search_deep_keys(self, search_text, k8n_dict, path):
        """Search deep for keys and get their values"""
        keys = []
        if isinstance(k8n_dict, dict):
            for key in k8n_dict:
                pathprop = path[:]
                pathprop.append(key)
                if key == search_text:
                    pathprop.append(k8n_dict[key])
                    keys.append(pathprop)
                    # pop the last element off for nesting of found elements for
                    # dict and list checks
                    pathprop = pathprop[:-1]
                if isinstance(k8n_dict[key], dict):
                    keys.extend(self._search_deep_keys(search_text, k8n_dict[key], pathprop))
                elif isinstance(k8n_dict[key], list):
                    for index, item in enumerate(k8n_dict[key]):
                        pathproparr = pathprop[:]
                        pathproparr.append(index)
                        keys.extend(self._search_deep_keys(search_text, item, pathproparr))
        elif isinstance(k8n_dict, list):
            for index, item in enumerate(k8n_dict):
                pathprop = path[:]
                pathprop.append(index)
                keys.extend(self._search_deep_keys(search_text, item, pathprop))

        return keys

def get_skipped_checks(entity_conf):
    skipped = []
    metadata = {}
    if not isinstance(entity_conf,dict):
        return skipped
    if entity_conf["kind"] == "containers" or entity_conf["kind"] == "initContainers":
        metadata = entity_conf["parent_metadata"]
    else:
        if "metadata" in entity_conf.keys():
            metadata = entity_conf["metadata"]
    if "annotations" in metadata.keys() and metadata["annotations"] is not None:
        for key in metadata["annotations"].keys():
            skipped_item = {}
            if "checkov.io/skip" in key or "bridgecrew.io/skip" in key:
                if "CKV_K8S" in metadata["annotations"][key]:
                    if "=" in metadata["annotations"][key]:
                        (skipped_item["id"], skipped_item["suppress_comment"]) = metadata["annotations"][key].split("=")
                    else:
                        skipped_item["id"] = metadata["annotations"][key]
                        skipped_item["suppress_comment"] = "No comment provided"
                    skipped.append(skipped_item)
                else:
                    logging.info("Parse of Annotation Failed for {}: {}".format(metadata["annotations"][key], entity_conf, indent=2))
                    continue
    return skipped

def _get_from_dict(data_dict, map_list):
    return reduce(operator.getitem, map_list, data_dict)


def _set_in_dict(data_dict, map_list, value):
    _get_from_dict(data_dict, map_list[:-1])[map_list[-1]] = value


def find_lines(node, kv):
    if isinstance(node, str):
        return node
    if isinstance(node, list):
        for i in node:
            for x in find_lines(i, kv):
                yield x
    elif isinstance(node, dict):
        if kv in node:
            yield node[kv]
        for j in node.values():
            for x in find_lines(j, kv):
                yield x





import re
from typing import Any

from checkov.terraform.variable_rendering.safe_eval_functions import SAFE_EVAL_DICT
# condition ? true_val : false_val -> (condition, true_val, false_val)
from checkov.terraform.parser_utils import find_var_blocks

CONDITIONAL_EXPR = r'([^\?]+)\?([^:]+)\:([^:]+)'

# {key1 = value1, key2 = value2, ...}
MAP_REGEX = r'\{(?:\s*[\S]+\s*\=\s*[\S]+\s*\,)*(?:\s*[\S]+\s*\=\s*[\S]+\s*)\}'

# {key:val}[key]
MAP_WITH_ACCESS = re.compile(r'(?P<d>{(?:.*?:.*?)+(,?:.*?:.*?)*})\s*(?P<access>\[[^\]]+\])')

LIST_PATTERN = r'(?P<d>\[([^\[\]]+?)+(\,[^\[\]]+?)*\])'

KEY_VALUE_REGEX = r'([\S]+)\s*\=\s*([\S]+)'

# %{ some_text }
DIRECTIVE_EXPR = r'\%\{([^\}]*)\}'

COMPARE_REGEX = re.compile(r'^(?P<a>.+)(?P<operator>==|!=|>=|>|<=|<|&&|\|\|)+(?P<b>.+)$')


def evaluate_terraform(input_str, keep_interpolations=True):
    evaluated_value = _try_evaluate(input_str)
    if type(evaluated_value) is not str:
        return input_str if callable(evaluated_value) else evaluated_value
    evaluated_value = evaluated_value.replace("\n", "")
    evaluated_value = evaluated_value.replace(",,", ",")
    if not keep_interpolations:
        evaluated_value = remove_interpolation(evaluated_value)
    evaluated_value = evaluate_map(evaluated_value)
    evaluated_value = evaluate_list_access(evaluated_value)
    evaluated_value = strip_double_quotes(evaluated_value)
    evaluated_value = evaluate_directives(evaluated_value)
    evaluated_value = evaluate_conditional_expression(evaluated_value)
    evaluated_value = evaluate_compare(evaluated_value)
    evaluated_value = evaluate_json_types(evaluated_value)
    second_evaluated_value = _try_evaluate(evaluated_value)

    return evaluated_value if callable(second_evaluated_value) else second_evaluated_value


def _try_evaluate(input_str):
    try:
        return eval(input_str, {"__builtins__": None}, SAFE_EVAL_DICT) # nosec
    except Exception:
        try:
            return eval(f'"{input_str}"', {"__builtins__": None}, SAFE_EVAL_DICT) # nosec
        except Exception:
            return input_str


def replace_string_value(original_str, str_to_replace, replaced_value, keep_origin=True):
    if original_str is None or type(original_str) not in (str, list):
        return original_str

    if type(original_str) is list:
        for i, item in enumerate(original_str):
            original_str[i] = replace_string_value(item, str_to_replace, replaced_value, keep_origin)
            if type(replaced_value) in [int, float, bool]:
                original_str[i] = evaluate_terraform(original_str[i])
            return original_str

    if str_to_replace not in original_str:
        return original_str if keep_origin else str_to_replace

    string_without_interpolation = remove_interpolation(original_str, str_to_replace)
    return string_without_interpolation.replace(str_to_replace, str(replaced_value)).replace(' ', '')


def remove_interpolation(original_str, var_to_clean: str=None):
    # get all variable references in string
    # remove from the string all ${} or '${}' occurrences
    var_blocks = find_var_blocks(original_str)
    var_blocks.reverse()
    for block in var_blocks:
        if block.full_str.startswith("${") and block.full_str.endswith("}") and (not var_to_clean or block.var_only == var_to_clean):
            full_str_start = original_str.find(block.full_str)
            full_str_end = full_str_start + len(block.full_str)
            if full_str_start > 0 and full_str_end <= len(original_str) - 2 and original_str[full_str_start-1] == "'" and original_str[full_str_start-1] == original_str[full_str_end] and "." in block.full_str:
                # checking if ${} is wrapped with '' like : '${}'
                original_str = original_str[:full_str_start-1] + block.full_str + original_str[full_str_end+1:]
            original_str = original_str.replace(block.full_str, block.var_only)
    return original_str


def strip_double_quotes(input_str):
    if input_str.startswith('"') and input_str.endswith('"'):
        input_str = input_str[1:-1]
    return input_str


def evaluate_conditional_expression(input_str: str) -> str:
    variable_ref = re.match(r"^\${(.*)}$", input_str)
    if variable_ref:
        input_str = variable_ref.groups()[0]

    condition = re.match(CONDITIONAL_EXPR, input_str)
    while condition:
        groups = condition.groups()
        if len(groups) != 3:
            return input_str
        evaluated_condition = evaluate_terraform(groups[0])
        condition_substr = input_str[condition.start():condition.end()]
        if convert_to_bool(evaluated_condition):
            true_val = str(evaluate_terraform(groups[1]))
            input_str = input_str.replace(condition_substr, true_val)
        else:
            false_val = str(evaluate_terraform(groups[2]))
            input_str = input_str.replace(condition_substr, false_val)
        condition = re.match(CONDITIONAL_EXPR, input_str)

    return input_str


def evaluate_compare(input_str):
    """
    :param input_str: string like "a && b" (supported operators: ==, != , <, <=, >, >=, && , ||)
    :return: evaluation of the expression
    """
    if type(input_str) is str and re.search(COMPARE_REGEX, input_str) and 'for' not in input_str:
        compare_parts = re.search(COMPARE_REGEX, input_str).groupdict()
        a = compare_parts.get('a')
        b = compare_parts.get('b')
        op = compare_parts.get('operator')
        if a and b and op:
            try:
                return apply_binary_op(evaluate_terraform(a), evaluate_terraform(b), op)
            except TypeError or SyntaxError:
                return input_str

    return input_str


def evaluate_json_types(input_str: Any) -> str:
    # https://www.terraform.io/docs/language/functions/jsonencode.html
    if isinstance(input_str, str) and input_str.startswith("jsonencode("):
        return input_str.replace("true", "True").replace("false", "False").replace("null", "None")

    return input_str


def apply_binary_op(a, b, operator):
    # apply the operator after verifying that a and b have the same type.
    operators = {
        '==': lambda a, b: a == b,
        '!=': lambda a, b: a != b,
        '>': lambda a, b: a > b,
        '>=': lambda a, b: a >= b,
        '<': lambda a, b: a < b,
        '<=': lambda a, b: a <= b,
        '&&': lambda a, b: a and b,
        '||': lambda a, b: a or b
    }
    type_a = type(a)
    type_b = type(b)

    if type_a != type_b:
        try:
            temp_b = type_a(b)
            if type_a == bool:
                temp_b = bool(convert_to_bool(b))
            return operators[operator](a, temp_b)
        except Exception:
            temp_a = type_b(a)
            if type_b == bool:
                temp_a = bool(convert_to_bool(a))
            return operators[operator](temp_a, b)
    else:
        return operators[operator](a, b)


def evaluate_directives(input_str):
    if re.search(DIRECTIVE_EXPR, input_str) is None:
        return input_str

    # replace `%{if <BOOL>}%{true_val}%{else}%{false_val}%{endif}` pattern with `<BOOL> ? true_val : false_val`
    matching_directives = re.findall(DIRECTIVE_EXPR, input_str)
    if len(matching_directives) == 3:
        if re.search(r'\bif\b', matching_directives[0]) and re.search(r'\belse\b', matching_directives[1]) and re.search(r'\bendif\b', matching_directives[2]):
            split_by_directives = re.split(DIRECTIVE_EXPR, input_str)
            edited_str = ''
            for part in split_by_directives:
                if re.search(r'\bif\b', part):
                    part = part.replace('if', '%{') + ' ? '
                    part = re.sub(r'\s', '', part)
                if re.search(r'\belse\b', part):
                    part = part.replace('else', ':')
                    part = re.sub(r'\s', '', part)
                if re.search(r'\bendif\b', part):
                    part = part.replace('endif', '}')
                    part = re.sub(r'\s', '', part)
                edited_str += part
            input_str = edited_str

    matching_directives = re.split(DIRECTIVE_EXPR, input_str)
    evaluated_string_parts = []
    for str_part in matching_directives:
        evaluated_string_parts.append(evaluate_terraform(str_part))
    return ''.join(evaluated_string_parts)


def evaluate_map(input_str):
    # first replace maps ":" with "="
    all_curly_brackets = find_brackets_pairs(input_str, "{", "}")
    for curly_match in all_curly_brackets:
        curly_start = curly_match["start"]
        curly_end = curly_match["end"]
        replaced_matching_map = input_str[curly_start:curly_end+1].replace("=", ":")
        input_str = input_str.replace(input_str[curly_start:curly_end+1], replaced_matching_map)

    # find map access like {a: b}[a] and extract the right value - b
    all_square_brackets = find_brackets_pairs(input_str, "[", "]")

    curr_square_match = 0
    for curly_match in all_curly_brackets:
        curly_start = curly_match["start"]
        curly_end = curly_match["end"]
        for i in range(curr_square_match, len(all_square_brackets)):
            curr_square_match = i
            square_match = all_square_brackets[i]
            square_start = square_match["start"]
            square_end = square_match["end"]
            if square_start > curly_end and (square_start == curly_end + 1 or all(c == ' ' for c in input_str[curly_end + 1:square_start])):
                origin_match_str = input_str[curly_start:square_end + 1]
                map_access = input_str[square_start+1:square_end]
                if not map_access.startswith('"') and not map_access.endswith('"'):
                    origin_match_str = origin_match_str.replace(f'[{map_access}]', f'["{map_access}"]')
                evaluated = _try_evaluate(origin_match_str)
                if evaluated:
                    input_str = input_str.replace(input_str[curly_start:square_end + 1], str(evaluated))
                    break

    return input_str


def convert_to_bool(bool_str):
    if bool_str in ['true', '"true"', 'True', '"True"', 1, '1']:
        return True
    elif bool_str in ['false', '"false"', 'False', '"False"', 0, '0']:
        return False
    else:
        return bool_str


def evaluate_list_access(input_str):
    # find list access like [a, b, c][0] and extract the right value - a

    all_square_brackets = find_brackets_pairs(input_str, "[", "]")
    prev_start = -1
    prev_end = -1
    for match in all_square_brackets:
        if (match["start"] == prev_end + 1 or all(c == ' ' for c in input_str[prev_end+1:match["start"]])) and prev_start != -1:
            curr_str = input_str[match["start"]+1:match["end"]]
            if curr_str.isnumeric():
                evaluated = _try_evaluate(input_str[prev_start:match["end"]+1])
                if evaluated:
                    input_str = input_str.replace(input_str[prev_start:match["end"]+1], str(evaluated))
        prev_start = match["start"]
        prev_end = match["end"]

    return input_str


def find_brackets_pairs(input_str, starting, closing):
    brackets_pairs = [-1] * len(input_str)
    unmatched_open = []

    for i, c in enumerate(input_str):
        if c == starting:
            unmatched_open.append(i)
        elif c == closing and len(unmatched_open) > 0:
            brackets_pairs[unmatched_open[-1]] = i
            unmatched_open = unmatched_open[:-1]

    all_brackets = []
    for start, end in enumerate(brackets_pairs):
        if end != -1:
            all_brackets.append({"start": start, "end": end})
    return all_brackets



import logging
import operator
import os
from functools import reduce
from checkov.common.util.type_forcers import force_list
from checkov.common.output.record import Record
from checkov.common.output.report import Report
from checkov.common.runners.base_runner import BaseRunner, filter_ignored_directories
from checkov.kubernetes.parser.parser import parse
from checkov.kubernetes.registry import registry
from checkov.runner_filter import RunnerFilter

K8_POSSIBLE_ENDINGS = [".yaml", ".yml", ".json"]


class Runner(BaseRunner):
    check_type = "kubernetes"

    def run(self, root_folder, external_checks_dir=None, files=None, runner_filter=RunnerFilter(), collect_skip_comments=True, helmChart=None):
        report = Report(self.check_type)
        definitions = {}
        definitions_raw = {}
        parsing_errors = {}
        files_list = []
        if external_checks_dir:
            for directory in external_checks_dir:
                registry.load_external_checks(directory)

        if files:
            for file in files:
                parse_result = parse(file)
                if parse_result:
                    (definitions[file], definitions_raw[file]) = parse_result

        if root_folder:
            for root, d_names, f_names in os.walk(root_folder):
                filter_ignored_directories(d_names)

                for file in f_names:
                    file_ending = os.path.splitext(file)[1]
                    if file_ending in K8_POSSIBLE_ENDINGS:
                        full_path = os.path.join(root, file)
                        if "/." not in full_path and file not in ['package.json','package-lock.json']:
                            # skip temp directories
                            files_list.append(full_path)

            for file in files_list:
                relative_file_path = f'/{os.path.relpath(file, os.path.commonprefix((root_folder, file)))}'
                parse_result = parse(file)
                if parse_result:
                    (definitions[relative_file_path], definitions_raw[relative_file_path]) = parse_result

        for k8_file in definitions.keys():

            # There are a few cases here. If -f was used, there could be a leading / because it's an absolute path,
            # or there will be no leading slash; root_folder will always be none.
            # If -d is used, root_folder will be the value given, and -f will start with a / (hardcoded above).
            # The goal here is simply to get a valid path to the file (which sls_file does not always give).
            if k8_file[0] == '/':
                path_to_convert = (root_folder + k8_file) if root_folder else k8_file
            else:
                path_to_convert = (os.path.join(root_folder, k8_file)) if root_folder else k8_file

            file_abs_path = os.path.abspath(path_to_convert)

            if definitions[k8_file]:
                for i in range(len(definitions[k8_file])):
                    if (not 'apiVersion' in definitions[k8_file][i].keys()) and (not 'kind' in definitions[k8_file][i].keys()):
                        continue
                    logging.debug("Template Dump for {}: {}".format(k8_file, definitions[k8_file][i], indent=2))

                    entity_conf = definitions[k8_file][i]

                    # Split out resources if entity kind is List
                    if entity_conf["kind"] == "List":
                        for item in entity_conf["items"]:
                            definitions[k8_file].append(item)

                for i in range(len(definitions[k8_file])):
                    if (not 'apiVersion' in definitions[k8_file][i].keys()) and (not 'kind' in definitions[k8_file][i].keys()):
                        continue
                    logging.debug("Template Dump for {}: {}".format(k8_file, definitions[k8_file][i], indent=2))

                    entity_conf = definitions[k8_file][i]

                    if entity_conf["kind"] == "List":
                        continue

                    # Skip entity without metadata["name"]
                    if entity_conf.get("metadata"):
                        if isinstance(entity_conf["metadata"], int) or not "name" in entity_conf["metadata"]:
                            continue
                    else:
                        continue

                    # Skip entity with parent (metadata["ownerReferences"]) in runtime
                    # We will alert in runtime only
                    if "ownerReferences" in entity_conf["metadata"] and \
                            entity_conf["metadata"]["ownerReferences"] is not None:
                        continue

                    # Append containers and initContainers to definitions list
                    for type in ["containers", "initContainers"]:
                        containers = []
                        if entity_conf["kind"] == "CustomResourceDefinition":
                            continue
                        containers = self._search_deep_keys(type, entity_conf, [])
                        if not containers:
                            continue
                        containers = containers.pop()
                        #containers.insert(0,entity_conf['kind'])
                        containerDef = {}
                        namespace = ""
                        if "namespace" in entity_conf["metadata"]:
                            namespace = entity_conf["metadata"]["namespace"]
                        else:
                            namespace = "default"
                        containerDef["containers"] = containers.pop()
                        if containerDef["containers"] is not None:
                            for cd in containerDef["containers"]:
                                i = containerDef["containers"].index(cd)
                                containerDef["containers"][i]["apiVersion"] = entity_conf["apiVersion"]
                                containerDef["containers"][i]["kind"] = type
                                containerDef["containers"][i]["parent"] = "{}.{}.{} (container {})".format(
                                    entity_conf["kind"], entity_conf["metadata"]["name"], namespace, str(i))
                                containerDef["containers"][i]["parent_metadata"] = entity_conf["metadata"]
                            definitions[k8_file].extend(containerDef["containers"])

                # Run for each definition included added container definitions
                for i in range(len(definitions[k8_file])):
                    if (not 'apiVersion' in definitions[k8_file][i].keys()) and (not 'kind' in definitions[k8_file][i].keys()):
                        continue
                    logging.debug("Template Dump for {}: {}".format(k8_file, definitions[k8_file][i], indent=2))

                    entity_conf = definitions[k8_file][i]

                    if entity_conf["kind"] == "List" or not entity_conf.get("kind"):
                        continue

                    if isinstance(entity_conf["kind"], int):
                        continue
                    # Skip entity without metadata["name"] or parent_metadata["name"]
                    if not any(x in entity_conf["kind"] for x in ["containers", "initContainers"]):
                        if entity_conf.get("metadata"):
                            if isinstance(entity_conf["metadata"], int) or not "name" in entity_conf["metadata"]:
                                continue
                        else:
                            continue

                    # Skip entity with parent (metadata["ownerReferences"]) in runtime
                    # We will alert in runtime only
                    if "metadata" in entity_conf:
                        if "ownerReferences" in entity_conf["metadata"] and \
                                entity_conf["metadata"]["ownerReferences"] is not None:
                            continue

                    # Skip Kustomization Templates (for now)
                    if entity_conf["kind"] == "Kustomization":
                        continue

                    skipped_checks = get_skipped_checks(entity_conf)

                    results = registry.scan(k8_file, entity_conf, skipped_checks, runner_filter)

                    # TODO refactor into context parsing
                    find_lines_result_list = list(find_lines(entity_conf, '__startline__'))
                    start_line = entity_conf["__startline__"]
                    end_line = entity_conf["__endline__"]

                    if start_line == end_line:
                        entity_lines_range = [start_line, end_line]
                        entity_code_lines = definitions_raw[k8_file][start_line - 1: end_line]
                    else:
                        entity_lines_range = [start_line, end_line - 1]
                        entity_code_lines = definitions_raw[k8_file][start_line - 1: end_line - 1]

                    # TODO? - Variable Eval Message!
                    variable_evaluations = {}

                    for check, check_result in results.items():
                        record = Record(check_id=check.id, check_name=check.name, check_result=check_result,
                                        code_block=entity_code_lines, file_path=k8_file,
                                        file_line_range=entity_lines_range,
                                        resource=check.get_resource_id(entity_conf), evaluations=variable_evaluations,
                                        check_class=check.__class__.__module__, file_abs_path=file_abs_path)
                        report.add_record(record=record)

        return report



    def _search_deep_keys(self, search_text, k8n_dict, path):
        """Search deep for keys and get their values"""
        keys = []
        if isinstance(k8n_dict, dict):
            for key in k8n_dict:
                pathprop = path[:]
                pathprop.append(key)
                if key == search_text:
                    pathprop.append(k8n_dict[key])
                    keys.append(pathprop)
                    # pop the last element off for nesting of found elements for
                    # dict and list checks
                    pathprop = pathprop[:-1]
                if isinstance(k8n_dict[key], dict):
                    keys.extend(self._search_deep_keys(search_text, k8n_dict[key], pathprop))
                elif isinstance(k8n_dict[key], list):
                    for index, item in enumerate(k8n_dict[key]):
                        pathproparr = pathprop[:]
                        pathproparr.append(index)
                        keys.extend(self._search_deep_keys(search_text, item, pathproparr))
        elif isinstance(k8n_dict, list):
            for index, item in enumerate(k8n_dict):
                pathprop = path[:]
                pathprop.append(index)
                keys.extend(self._search_deep_keys(search_text, item, pathprop))

        return keys

def get_skipped_checks(entity_conf):
    skipped = []
    metadata = {}
    if not isinstance(entity_conf,dict):
        return skipped
    if entity_conf["kind"] == "containers" or entity_conf["kind"] == "initContainers":
        metadata = entity_conf["parent_metadata"]
    else:
        if "metadata" in entity_conf.keys():
            metadata = entity_conf["metadata"]
    if "annotations" in metadata.keys() and metadata["annotations"] is not None:
        if isinstance(metadata["annotations"], dict):
            metadata["annotations"] = force_list(metadata["annotations"])
        for annotation in metadata["annotations"]:
            if not isinstance(annotation, dict):
                logging.debug( f"Parse of Annotation Failed for {annotation}: {entity_conf}")
                continue
            for key in annotation:
                skipped_item = {}
                if "checkov.io/skip" in key or "bridgecrew.io/skip" in key:
                    if "CKV_K8S" in annotation[key]:
                        if "=" in annotation[key]:
                            (skipped_item["id"], skipped_item["suppress_comment"]) = annotation[key].split("=")
                        else:
                            skipped_item["id"] = annotation[key]
                            skipped_item["suppress_comment"] = "No comment provided"
                        skipped.append(skipped_item)
                    else:
                        logging.debug("Parse of Annotation Failed for {}: {}".format(metadata["annotations"][key], entity_conf, indent=2))
                        continue
    return skipped

def _get_from_dict(data_dict, map_list):
    return reduce(operator.getitem, map_list, data_dict)


def _set_in_dict(data_dict, map_list, value):
    _get_from_dict(data_dict, map_list[:-1])[map_list[-1]] = value


def find_lines(node, kv):
    if isinstance(node, str):
        return node
    if isinstance(node, list):
        for i in node:
            for x in find_lines(i, kv):
                yield x
    elif isinstance(node, dict):
        if kv in node:
            yield node[kv]
        for j in node.values():
            for x in find_lines(j, kv):
                yield x





from checkov.common.models.enums import CheckResult
from checkov.runner_filter import RunnerFilter


class BaseRegistry:
    def __init__(self, parser):
        self.checks = []
        self.parser = parser

    def load_checks(self):
        raise NotImplementedError

    def run_checks(self, graph_connector, runner_filter: RunnerFilter):
        check_results = {}
        for check in self.checks:
            if runner_filter.checks and check.id not in runner_filter.checks \
                    or runner_filter.skip_checks and check.id in runner_filter.skip_checks:
                continue
            passed, failed = check.run(graph_connector)
            check_result = self._process_check_result(passed, [], CheckResult.PASSED)
            check_result = self._process_check_result(failed, check_result, CheckResult.FAILED)
            check_results[check] = check_result
        return check_results

    @staticmethod
    def _process_check_result(results, processed_results, result):
        for vertex in results:
            processed_results.append({'result': result, 'entity': vertex})
        return processed_results



from checkov.common.models.enums import CheckCategories, CheckResult
from checkov.terraform.checks.resource.base_resource_negative_value_check import BaseResourceNegativeValueCheck


class FunctionAppDisallowCORS(BaseResourceNegativeValueCheck):
    def __init__(self):
        name = "Ensure function apps are not accessible from all regions"
        id = "CKV_AZURE_62"
        supported_resources = ['azurerm_function_app']
        categories = [CheckCategories.GENERAL_SECURITY]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources, missing_attribute_result=CheckResult.PASSED)

    def get_inspected_key(self):
        return 'site_config/[0]/cors/[0]/allowed_origins'

    def get_forbidden_values(self):
        return [['*']]


check = FunctionAppDisallowCORS()



from checkov.common.models.enums import CheckResult, CheckCategories
from checkov.arm.base_resource_check import BaseResourceCheck
from checkov.common.util.type_forcers import force_int

# https://docs.microsoft.com/en-us/azure/templates/microsoft.insights/2016-03-01/logprofiles

class MonitorLogProfileRetentionDays(BaseResourceCheck):
    def __init__(self):
        name = "Ensure that Activity Log Retention is set 365 days or greater"
        id = "CKV_AZURE_37"
        supported_resources = ['microsoft.insights/logprofiles']
        categories = [CheckCategories.LOGGING]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf):
        if "properties" in conf:
            if "retentionPolicy" in conf["properties"]:
                if "enabled" in conf["properties"]["retentionPolicy"]:
                    if str(conf["properties"]["retentionPolicy"]["enabled"]).lower() == "true":
                        if "days" in conf["properties"]["retentionPolicy"]:
                            if force_int(conf["properties"]["retentionPolicy"]["days"]) >= 365 or \
                                    force_int(conf["properties"]["retentionPolicy"]["days"]) == 0:
                                return CheckResult.PASSED
        return CheckResult.FAILED

check = MonitorLogProfileRetentionDays()
